'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const api = require('@previewjs/api');
const typeAnalyzer = require('@previewjs/type-analyzer');
const cookieParser = require('cookie-parser');
const express = require('express');
const fs$1 = require('fs-extra');
const getPort = require('get-port');
const path = require('path');
const config = require('@previewjs/config');
const glob = require('glob');
const util = require('util');
const fs = require('fs');
const envPaths = require('env-paths');
const vfs = require('@previewjs/vfs');
const assertNever = require('assert-never');
const httpTerminator = require('http-terminator');
const recrawl = require('recrawl');
const fakeExportedTypesPlugin = require('rollup-plugin-friendly-type-imports');
const tsconfigLoader_js = require('tsconfig-paths/lib/tsconfig-loader.js');
const vite = require('vite');
const viteTsconfigPaths = require('vite-tsconfig-paths');
const url = require('url');
const acorn = require('acorn');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e["default"] : e; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  const n = Object.create(null);
  if (e) {
    for (const k in e) {
      n[k] = e[k];
    }
  }
  n["default"] = e;
  return n;
}

const cookieParser__default = /*#__PURE__*/_interopDefaultLegacy(cookieParser);
const express__default = /*#__PURE__*/_interopDefaultLegacy(express);
const fs__default$1 = /*#__PURE__*/_interopDefaultLegacy(fs$1);
const getPort__default = /*#__PURE__*/_interopDefaultLegacy(getPort);
const path__default = /*#__PURE__*/_interopDefaultLegacy(path);
const glob__default = /*#__PURE__*/_interopDefaultLegacy(glob);
const fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
const envPaths__default = /*#__PURE__*/_interopDefaultLegacy(envPaths);
const assertNever__default = /*#__PURE__*/_interopDefaultLegacy(assertNever);
const fakeExportedTypesPlugin__default = /*#__PURE__*/_interopDefaultLegacy(fakeExportedTypesPlugin);
const vite__namespace = /*#__PURE__*/_interopNamespace(vite);
const viteTsconfigPaths__default = /*#__PURE__*/_interopDefaultLegacy(viteTsconfigPaths);

function getCacheDir(rootDirPath) {
  try {
    const { version } = JSON.parse(fs__default.readFileSync(path__default.resolve(__dirname, "..", "..", "package.json"), "utf8"));
    return path__default.resolve(rootDirPath, "node_modules", ".previewjs", `v${version}`);
  } catch (e) {
    throw new Error(`Unable to detect @previewjs/core version.`);
  }
}

async function analyzeProject(workspace, options = {}) {
  const cacheFilePath = path__default.join(getCacheDir(workspace.rootDirPath), "components.json");
  useCache:
    if (!options.forceRefresh && await fs__default$1.pathExists(cacheFilePath)) {
      const stat = await fs__default$1.stat(cacheFilePath);
      if (stat.mtimeMs < Date.now() - 24 * 3600 * 1e3) {
        break useCache;
      }
      return {
        components: JSON.parse(await fs__default$1.readFile(cacheFilePath, "utf8")),
        cached: true
      };
    }
  const components = await analyzeProjectCore(workspace);
  await fs__default$1.mkdirp(path__default.dirname(cacheFilePath));
  await fs__default$1.writeFile(cacheFilePath, JSON.stringify(components));
  return { components, cached: false };
}
async function analyzeProjectCore(workspace) {
  const absoluteFilePaths = await findFiles(workspace.rootDirPath, "**/*.@(js|jsx|ts|tsx|svelte|vue)");
  const components = {};
  const found = await workspace.frameworkPlugin.detectComponents(workspace.typeAnalyzer, absoluteFilePaths);
  for (const component of found) {
    const filePath = path__default.relative(workspace.rootDirPath, component.absoluteFilePath);
    const fileComponents = components[filePath] || (components[filePath] = []);
    fileComponents.push({
      componentName: component.name,
      exported: component.exported
    });
  }
  return components;
}
async function findFiles(rootDirPath, pattern) {
  const config$1 = await config.readConfig(rootDirPath);
  const files = await util.promisify(glob__default)(pattern, {
    ignore: ["**/node_modules/**", ...config$1.exclude || []],
    cwd: rootDirPath,
    nodir: true,
    absolute: true,
    follow: false
  });
  return files.filter((f) => f.startsWith(rootDirPath.replace(/\\/g, "/") + "/"));
}

async function computeProps({
  component
}) {
  const result = await component.analyze();
  return {
    types: {
      props: result.propsType,
      all: result.types
    },
    args: [...result.providedArgs]
  };
}

var __accessCheck = (obj, member, msg) => {
  if (!member.has(obj))
    throw TypeError("Cannot " + msg);
};
var __privateAdd = (obj, member, value) => {
  if (member.has(obj))
    throw TypeError("Cannot add the same private member more than once");
  member instanceof WeakSet ? member.add(obj) : member.set(obj, value);
};
var __privateMethod = (obj, member, method) => {
  __accessCheck(obj, member, "access private method");
  return method;
};
var _read, read_fn, _update, update_fn;
class LocalFilePersistedStateManager {
  constructor(absoluteFilePath = path__default.join(envPaths__default("previewjs").config, "state.json")) {
    this.absoluteFilePath = absoluteFilePath;
    __privateAdd(this, _read);
    __privateAdd(this, _update);
    this.get = async () => {
      const state = await __privateMethod(this, _read, read_fn).call(this);
      return state || __privateMethod(this, _update, update_fn).call(this, {});
    };
    this.update = async (partialState) => {
      return __privateMethod(this, _update, update_fn).call(this, partialState);
    };
  }
}
_read = new WeakSet();
read_fn = async function() {
  try {
    const content = await fs__default$1.readFile(this.absoluteFilePath, "utf8");
    return JSON.parse(content);
  } catch {
    return null;
  }
};
_update = new WeakSet();
update_fn = async function(partialState) {
  const state = {
    ...await (__privateMethod(this, _read, read_fn).call(this) || {}),
    ...partialState
  };
  await fs__default$1.ensureDir(path__default.dirname(this.absoluteFilePath));
  await fs__default$1.writeFile(this.absoluteFilePath, JSON.stringify(state, null, 2), "utf8");
  return state;
};

class Server {
  constructor(options) {
    this.serverTerminator = null;
    this.server = null;
    const app = express__default();
    app.use((_req, res, next) => {
      res.setHeader("Cache-Control", "max-age=0, must-revalidate");
      next();
    });
    for (const middleware of options.middlewares) {
      app.use(middleware);
    }
    this.app = app;
  }
  async start(port) {
    return new Promise((resolve) => {
      const server = this.server = this.app.listen(port, () => {
        console.log(`Preview.js Server running at http://localhost:${port}.`);
        resolve(server);
      });
      this.serverTerminator = httpTerminator.createHttpTerminator({
        server: this.server,
        gracefulTerminationTimeout: 0
      });
    });
  }
  async stop() {
    if (this.serverTerminator) {
      await this.serverTerminator.terminate();
      this.serverTerminator = null;
    }
    if (this.server) {
      const server = this.server;
      await new Promise((resolve) => server.close(resolve));
      this.server = null;
    }
    console.log(`Preview.js Server stopped.`);
  }
}

const COMPONENT_LOADER_MODULE = "/@component-loader.js";
function componentLoaderPlugin(options) {
  return {
    name: "previewjs:component-loader",
    resolveId: async function(id) {
      if (id.startsWith(COMPONENT_LOADER_MODULE)) {
        return id;
      }
      return null;
    },
    load: async function(id) {
      if (!id.startsWith(COMPONENT_LOADER_MODULE)) {
        return null;
      }
      const params = new url.URLSearchParams(id.split("?")[1] || "");
      return generateComponentLoaderModule(params, options.config.wrapper);
    }
  };
}
function generateComponentLoaderModule(urlParams, wrapper) {
  const filePath = urlParams.get("p");
  const componentName = urlParams.get("c");
  if (filePath === null || componentName === null) {
    throw new Error(`Invalid use of /@component-loader.js module`);
  }
  const componentModuleId = `/${filePath.replace(/\\/g, "/")}`;
  return `import { updateComponent } from '/__previewjs_internal__/update-component';
import { load } from '/__previewjs_internal__/renderer/index';

let counter = 0;
export async function refresh() {
  const currentCounter = ++counter;
  let loadingError = null;
  ${wrapper ? `
  let wrapperModulePromise;
  if (import.meta.hot.data.preloadedWrapperModule) {
    wrapperModulePromise = Promise.resolve(import.meta.hot.data.preloadedWrapperModule);
  } else {
    wrapperModulePromise = import("/${wrapper.path}");
  }
  const wrapperModule = await wrapperModulePromise.catch(e => {
    console.error(e);
    loadingError = e.stack || e.message || null;
    return null;
  });
  ` : `
  const wrapperModule = null;
  `}
  let componentModulePromise;
  if (import.meta.hot.data.preloadedComponentModule) {
    componentModulePromise = Promise.resolve(import.meta.hot.data.preloadedComponentModule);
  } else {
    componentModulePromise = import("${componentModuleId}");
  }
  const componentModule = await componentModulePromise.catch(e => {
    console.error(e);
    loadingError = e.stack || e.message || null;
    return null;
  });
  if (currentCounter !== counter) {
    // Abort to avoid double rendering.
    return;
  }
  await updateComponent({
    wrapperModule,
    wrapperName: ${JSON.stringify(wrapper?.componentName || null)},
    componentModule,
    componentFilePath: ${JSON.stringify(filePath)},
    componentName: ${JSON.stringify(componentName)},
    loadingError,
    load,
  })
}

import.meta.hot.accept();

${wrapper ? `
import.meta.hot.accept(["${wrapper.path}"], ([wrapperModule]) => {
  if (wrapperModule) {
    import.meta.hot.data.preloadedWrapperModule = wrapperModule;
    refresh();
  }
});
` : ``}

import.meta.hot.accept(["${componentModuleId}"], ([componentModule]) => {
  if (componentModule) {
    import.meta.hot.data.preloadedComponentModule = componentModule;
    refresh();
  }
});
`;
}

const extensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
const importRegexp = /import\s+([\S]+)\s+from\s+('|")([\S]+)\.(css|less|sass|scss|styl|stylus|postcss)(\?[\S]*)?&?('|")/g;
function cssModulesWithoutSuffixPlugin() {
  return {
    name: "previewjs:css-modules-without-suffix",
    async transform(code, id) {
      id = id.split("?")[0];
      if (!extensions.has(path__default.extname(id))) {
        return;
      }
      return code.replace(importRegexp, 'import $1 from "$3.$4?$5&.module.css"');
    }
  };
}

const jsExtensions$1 = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function exportToplevelPlugin() {
  return {
    name: "previewjs:export-toplevel",
    enforce: "post",
    transform: async function(code, id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      const extension = path__default.extname(id);
      if (!jsExtensions$1.has(extension)) {
        return null;
      }
      const topLevelEntityNames = findTopLevelEntityNames(id, code);
      return `${code};
          export {
          ${topLevelEntityNames.map((c) => `${c} as __previewjs__${c},`).join("")}
        }`;
    }
  };
}
function findTopLevelEntityNames(filePath, source) {
  let parsed;
  try {
    parsed = acorn.Parser.parse(source, {
      ecmaVersion: "latest",
      sourceType: "module"
    });
  } catch (e) {
    throw new Error(`Unable to parse ${filePath} with Acorn: ${e}`);
  }
  const topLevelEntityNames = [];
  for (const statement of parsed.body || []) {
    if (statement.type === "VariableDeclaration") {
      for (const declaration of statement.declarations) {
        if (declaration.type === "VariableDeclarator") {
          addIfIdentifier(topLevelEntityNames, declaration.id);
        }
      }
    }
    if (statement.type === "FunctionDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ClassDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ExportDefaultDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
    }
    if (statement.type === "ExportNamedDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
      if (statement.declaration?.declarations) {
        for (const declarator of statement.declaration.declarations) {
          addIfIdentifier(topLevelEntityNames, declarator.id);
        }
      }
    }
  }
  return topLevelEntityNames;
}
function addIfIdentifier(array, id) {
  if (id && id.type === "Identifier") {
    const name = id.name;
    if (name.endsWith("_default")) {
      return;
    }
    array.push(name);
  }
}

const VIRTUAL_PREFIX = `/@previewjs-virtual:`;
const jsExtensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function virtualPlugin(options) {
  const { reader, rootDirPath } = options;
  return {
    name: "previewjs:virtual-fs",
    resolveId: async function(id, importer) {
      const virtualImporter = importer?.startsWith(VIRTUAL_PREFIX) || false;
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      const extension = path__default.extname(id);
      let absoluteId;
      if (path__default.isAbsolute(id)) {
        absoluteId = id;
      } else {
        if (!importer) {
          return null;
        }
        if (virtualImporter) {
          importer = importer.slice(VIRTUAL_PREFIX.length);
        }
        if (extension && !jsExtensions.has(extension) && extension !== ".svg") {
          return path__default.join(path__default.dirname(importer), id);
        }
        absoluteId = path__default.join(path__default.dirname(importer), id);
      }
      const resolved = await resolveAbsoluteModuleId(absoluteId);
      if (resolved) {
        const [absoluteFilePath] = resolved;
        if (await fs__default$1.pathExists(absoluteFilePath) && !virtualImporter) {
          return null;
        }
        if (!absoluteId.endsWith(extension)) {
          absoluteId += extension;
        }
        const resolvedId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
        return resolvedId;
      }
      return null;
    },
    load: async function(id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      const resolved = await resolveAbsoluteModuleId(id);
      if (!resolved) {
        return null;
      }
      const [absoluteFilePath, entry] = resolved;
      if (entry.kind !== "file") {
        console.error(`Unable to read from ${absoluteFilePath}`);
        return null;
      }
      let source = await entry.read();
      const fileExtension = path__default.extname(absoluteFilePath);
      if (!jsExtensions.has(fileExtension)) {
        return source;
      }
      const moduleExtension = path__default.extname(id);
      return {
        code: moduleExtension === "" || moduleExtension === ".js" ? (await vite.transformWithEsbuild(source, absoluteFilePath, {
          loader: "tsx",
          format: "esm",
          sourcefile: path__default.relative(rootDirPath, absoluteFilePath),
          ...options.esbuildOptions
        })).code : source
      };
    },
    handleHotUpdate: async function(context) {
      const moduleGraph = options.moduleGraph();
      if (!moduleGraph) {
        return;
      }
      const filePath = path__default.relative(rootDirPath, context.file);
      const absoluteFilePath = path__default.join(rootDirPath, filePath);
      const entry = await reader.read(absoluteFilePath);
      if (!entry || entry.kind !== "file") {
        return;
      }
      const virtualModuleId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
      const node = moduleGraph.getModuleById(virtualModuleId);
      return node && [node];
    }
  };
  async function resolveAbsoluteModuleId(id) {
    return await resolveBaseFilePath(id) || await resolveBaseFilePath(path__default.join(rootDirPath, id));
  }
  async function resolveBaseFilePath(baseFilePath) {
    for (const suffix of [
      "",
      ...jsExtensions,
      ...[...jsExtensions].map((extension) => `${path__default.sep}index${extension}`)
    ]) {
      const absoluteFilePath = `${baseFilePath}${suffix}`;
      const entry = await reader.read(absoluteFilePath);
      if (entry !== null) {
        return [absoluteFilePath, entry];
      }
    }
    return null;
  }
}

class ViteManager {
  constructor(options) {
    this.options = options;
    this.lastPingTimestamp = 0;
    const router = express__default.Router();
    router.get("/preview/", async (req, res) => {
      const template = await fs__default$1.readFile(this.options.shadowHtmlFilePath, "utf-8");
      if (!this.viteServer) {
        res.status(404).end(`Vite is not running.`);
        return;
      }
      res.status(200).set({ "Content-Type": "text/html" }).end(await this.viteServer.transformIndexHtml(req.originalUrl, template));
    });
    router.use("/ping", async (req, res) => {
      this.lastPingTimestamp = Date.now();
      res.json(JSON.stringify({
        pong: "match!"
      }));
    });
    router.use(async (req, res, next) => {
      const waitSeconds = 60;
      const waitUntil = Date.now() + waitSeconds * 1e3;
      while (!this.viteServer) {
        await new Promise((resolve) => setTimeout(resolve, 50));
        if (Date.now() > waitUntil) {
          throw new Error(`Vite server is not running after ${waitSeconds} seconds.`);
        }
      }
      this.viteServer.middlewares(req, res, next);
    });
    this.middleware = router;
  }
  async start(server, port) {
    const typeScriptConfigFilePaths = await recrawl.recrawl({
      only: ["tsconfig.json"],
      skip: ["node_modules", ".git"]
    })(this.options.rootDirPath);
    const validTypeScriptFilePaths = [];
    for (const configFilePath of typeScriptConfigFilePaths) {
      try {
        tsconfigLoader_js.loadTsconfig(configFilePath);
        validTypeScriptFilePaths.push(configFilePath);
      } catch (e) {
        console.warn(`Encountered an invalid tsconfig.json file, ignoring: ${configFilePath}`);
      }
    }
    let tsInferredAlias = {};
    if (typeScriptConfigFilePaths.includes("tsconfig.json")) {
      const config = tsconfigLoader_js.loadTsconfig("tsconfig.json");
      if (config?.compilerOptions?.baseUrl && config?.compilerOptions?.paths) {
        const { baseUrl, paths } = config.compilerOptions;
        for (const [match, mapping] of Object.entries(paths)) {
          const firstMapping = mapping[0];
          if (!firstMapping) {
            continue;
          }
          const matchNoWildcard = match.endsWith("/*") ? match.slice(0, match.length - 2) : match;
          const firstMappingNoWildcard = firstMapping.endsWith("/*") ? firstMapping.slice(0, firstMapping.length - 2) : firstMapping;
          tsInferredAlias[matchNoWildcard] = path__default.join(this.options.rootDirPath, baseUrl, firstMappingNoWildcard);
        }
      }
    }
    const alias = {
      ...tsInferredAlias,
      ...this.options.config.alias,
      ...this.options.config.vite?.resolve?.alias
    };
    const defaultLogger = vite__namespace.createLogger(this.options.logLevel);
    const frameworkPluginViteConfig = this.options.frameworkPlugin.viteConfig({
      ...this.options.config,
      alias
    });
    const vitePlugins = [
      viteTsconfigPaths__default({
        root: this.options.rootDirPath,
        projects: validTypeScriptFilePaths
      }),
      virtualPlugin({
        reader: this.options.reader,
        rootDirPath: this.options.rootDirPath,
        moduleGraph: () => this.viteServer?.moduleGraph || null,
        esbuildOptions: frameworkPluginViteConfig.esbuild || {}
      }),
      exportToplevelPlugin(),
      fakeExportedTypesPlugin__default({
        readFile: (absoluteFilePath) => this.options.reader.read(absoluteFilePath).then((entry) => {
          if (entry?.kind !== "file") {
            return null;
          }
          return entry.read();
        })
      }),
      cssModulesWithoutSuffixPlugin(),
      componentLoaderPlugin({
        config: this.options.config
      }),
      ...frameworkPluginViteConfig.plugins || [],
      ...this.options.config.vite?.plugins || []
    ];
    const plugins = vitePlugins.flat().map((plugin) => {
      if (!plugin || Array.isArray(plugin) || !plugin.handleHotUpdate) {
        return plugin;
      }
      const handleHotUpdate = plugin.handleHotUpdate.bind(plugin);
      return {
        ...plugin,
        handleHotUpdate: async (ctx) => {
          return handleHotUpdate({
            ...ctx,
            read: async () => {
              const entry = await this.options.reader.read(ctx.file);
              if (entry?.kind !== "file") {
                return ctx.read();
              }
              return entry.read();
            }
          });
        }
      };
    });
    this.viteServer = await vite__namespace.createServer({
      ...frameworkPluginViteConfig,
      ...this.options.config.vite,
      configFile: false,
      root: this.options.rootDirPath,
      base: "/preview/",
      server: {
        middlewareMode: true,
        hmr: {
          overlay: false,
          server,
          clientPort: port,
          ...typeof this.options.config.vite?.server?.hmr === "object" ? this.options.config.vite?.server?.hmr : {}
        },
        ...this.options.config.vite?.server
      },
      customLogger: {
        info: defaultLogger.info,
        warn: defaultLogger.warn,
        error: defaultLogger.error,
        warnOnce: defaultLogger.warnOnce,
        clearScreen: () => {
        },
        hasWarned: defaultLogger.hasWarned,
        hasErrorLogged: defaultLogger.hasErrorLogged
      },
      clearScreen: false,
      cacheDir: this.options.config.vite?.cacheDir || this.options.cacheDir,
      publicDir: this.options.config.vite?.publicDir || this.options.config.publicDir,
      plugins,
      define: {
        __filename: void 0,
        __dirname: void 0,
        ...frameworkPluginViteConfig.define,
        ...this.options.config.vite?.define
      },
      resolve: {
        ...this.options.config.vite?.resolve,
        alias: {
          "~": "",
          "@": "",
          ...alias,
          ...frameworkPluginViteConfig.resolve?.alias
        }
      }
    });
  }
  async stop() {
    if (this.viteServer) {
      await this.viteServer.close();
      delete this.viteServer;
    }
  }
  getConfig() {
    return this.options.config;
  }
  getLastPingTimestamp() {
    return this.lastPingTimestamp;
  }
  triggerReload(absoluteFilePath) {
    if (!this.viteServer) {
      return;
    }
    for (const onChange of this.viteServer.watcher.listeners("change")) {
      onChange(absoluteFilePath);
    }
  }
  async renderIndexHtml(originalUrl) {
    if (!this.viteServer) {
      throw new Error(`Vite server is not running.`);
    }
    const template = await fs__default$1.readFile(this.options.shadowHtmlFilePath, "utf-8");
    return await this.viteServer.transformIndexHtml(originalUrl, template);
  }
}

const POSTCSS_CONFIG_FILE = ["postcss.config.js", ".postcssrc.js"];
const FILES_REQUIRING_RESTART = /* @__PURE__ */ new Set([
  config.PREVIEW_CONFIG_NAME,
  "jsconfig.json",
  "tsconfig.json",
  "package.json",
  "package-lock.json",
  "pnpm-lock.yaml",
  "yarn.lock",
  ...POSTCSS_CONFIG_FILE
]);
const SHUTDOWN_CHECK_INTERVAL = 3e3;
const SHUTDOWN_AFTER_INACTIVITY = 1e4;
class Previewer {
  constructor(options) {
    this.options = options;
    this.appServer = null;
    this.viteManager = null;
    this.status = { kind: "stopped" };
    this.shutdownCheckInterval = null;
    this.disposeObserver = null;
    this.config = null;
    this.onFileChangeListener = {
      onChange: (absoluteFilePath, info) => {
        absoluteFilePath = path__default.resolve(absoluteFilePath);
        const wrapperPath = this.config ? this.configWithWrapper(this.config).wrapper?.path : null;
        if (!info.virtual && (FILES_REQUIRING_RESTART.has(path__default.basename(absoluteFilePath)) || wrapperPath && absoluteFilePath === path__default.resolve(this.options.rootDirPath, wrapperPath))) {
          if (this.status.kind === "starting" || this.status.kind === "started") {
            const port = this.status.port;
            console.log("New dependencies were detected. Restarting...");
            this.stop().then(async () => {
              await this.start(async () => port);
            }).catch(console.error);
          }
          return;
        }
        if (info.virtual) {
          this.viteManager?.triggerReload(absoluteFilePath);
        } else if (this.options.onFileChanged) {
          this.options.onFileChanged(absoluteFilePath);
        }
      }
    };
    this.transformingReader = vfs.createStackedReader([
      options.reader,
      vfs.createFileSystemReader({
        mapping: {
          from: options.frameworkPlugin.previewDirPath,
          to: path__default.join(options.rootDirPath, "__previewjs_internal__", "renderer")
        },
        watch: false
      }),
      vfs.createFileSystemReader({
        mapping: {
          from: options.previewDirPath,
          to: options.rootDirPath
        },
        watch: false
      })
    ]);
  }
  async start(allocatePort) {
    let port;
    const statusBeforeStart = this.status;
    switch (statusBeforeStart.kind) {
      case "starting":
        port = statusBeforeStart.port;
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          console.error(e);
          this.status = {
            kind: "stopped"
          };
          await this.startFromStopped(port);
        }
        break;
      case "started":
        if (this.shutdownCheckInterval) {
          clearInterval(this.shutdownCheckInterval);
          this.shutdownCheckInterval = null;
        }
        port = statusBeforeStart.port;
        break;
      case "stopping":
        port = statusBeforeStart.port;
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          console.error(e);
          this.status = {
            kind: "stopped"
          };
        }
        await this.start(async () => port);
        break;
      case "stopped":
        port = await allocatePort();
        await this.startFromStopped(port);
        break;
      default:
        throw assertNever__default(statusBeforeStart);
    }
    return port;
  }
  async startFromStopped(port) {
    this.status = {
      kind: "starting",
      port,
      promise: (async () => {
        process.chdir(this.options.rootDirPath);
        this.config = await config.readConfig(this.options.rootDirPath);
        this.appServer = new Server({
          middlewares: [
            ...this.options.middlewares || [],
            (req, res, next) => {
              this.viteManager?.middleware(req, res, next);
            }
          ]
        });
        if (this.transformingReader.observe) {
          this.disposeObserver = await this.transformingReader.observe(this.options.rootDirPath);
        }
        this.transformingReader.listeners.add(this.onFileChangeListener);
        this.viteManager = new ViteManager({
          rootDirPath: this.options.rootDirPath,
          shadowHtmlFilePath: path__default.join(this.options.previewDirPath, "index.html"),
          reader: this.transformingReader,
          cacheDir: path__default.join(getCacheDir(this.options.rootDirPath), "vite"),
          config: this.configWithWrapper(this.config),
          logLevel: this.options.logLevel,
          frameworkPlugin: this.options.frameworkPlugin
        });
        const server = await this.appServer.start(port);
        await this.viteManager.start(server, port);
        this.status = {
          kind: "started",
          port
        };
      })()
    };
    await this.status.promise;
  }
  async stop(options = {}) {
    if (this.status.kind !== "started") {
      return;
    }
    if (options.onceUnused) {
      if (!await this.shutdownCheck()) {
        this.shutdownCheckInterval = setInterval(() => {
          this.shutdownCheck().catch(console.error);
        }, SHUTDOWN_CHECK_INTERVAL);
      }
    } else {
      await this.stopNow();
    }
  }
  configWithWrapper(config) {
    if (!config.wrapper) {
      const defaultWrapperPath = this.options.frameworkPlugin.defaultWrapperPath;
      if (fs$1.pathExistsSync(path__default.join(this.options.rootDirPath, defaultWrapperPath))) {
        config.wrapper = {
          path: defaultWrapperPath
        };
      }
    }
    return config;
  }
  async shutdownCheck() {
    const lastPingTimestamp = this.viteManager?.getLastPingTimestamp() || 0;
    if (lastPingTimestamp + SHUTDOWN_AFTER_INACTIVITY > Date.now()) {
      return false;
    }
    if (this.shutdownCheckInterval) {
      clearInterval(this.shutdownCheckInterval);
      this.shutdownCheckInterval = null;
    }
    await this.stopNow();
    return true;
  }
  async stopNow() {
    if (this.status.kind === "starting") {
      await this.status.promise;
    }
    if (this.status.kind !== "started") {
      return;
    }
    this.transformingReader.listeners.remove(this.onFileChangeListener);
    this.status = {
      kind: "stopping",
      port: this.status.port,
      promise: (async () => {
        if (this.disposeObserver) {
          await this.disposeObserver();
          this.disposeObserver = null;
        }
        if (this.viteManager) {
          await this.viteManager.stop();
          this.viteManager = null;
        }
        if (this.appServer) {
          await this.appServer.stop();
          this.appServer = null;
        }
        this.status = {
          kind: "stopped"
        };
      })()
    };
    await this.status.promise;
  }
}

class ApiRouter {
  constructor() {
    this.handlers = /* @__PURE__ */ new Map();
    this.registerEndpoint = (endpoint, handler) => {
      if (this.handlers.has(endpoint.path)) {
        throw new Error(`Multiple handlers registered for endpoint '${endpoint.path}'`);
      }
      this.handlers.set(endpoint.path, handler);
    };
  }
  async handle(path, request, expressRequest, expressResponse) {
    const handler = this.handlers.get(path);
    if (!handler) {
      return {
        kind: "error",
        message: `No handler is registered for path '${path}'`
      };
    }
    try {
      const response = await handler(request, expressRequest, expressResponse);
      return {
        kind: "success",
        response
      };
    } catch (e) {
      console.error(`Handler ${path} failed`, e);
      return {
        kind: "error",
        message: e.message
      };
    }
  }
}

function generateComponentId(options) {
  return `${options.currentFilePath.replace(/\\/g, "/")}:${options.siblingFileName ? `${options.siblingFileName}/` : ""}${options.name}`;
}

async function loadPreviewEnv({
  rootDirPath,
  setupEnvironment,
  frameworkPluginFactories
}) {
  const dependencies = await extractPackageDependencies(rootDirPath);
  for (const candidate of frameworkPluginFactories || []) {
    if (await candidate.isCompatible(dependencies)) {
      return {
        previewEnv: await setupEnvironment({ rootDirPath }),
        frameworkPlugin: await candidate.create()
      };
    }
  }
  return null;
}
async function extractPackageDependencies(rootDirPath) {
  const packageJsonPath = path__default.join(rootDirPath, "package.json");
  if (!await fs__default$1.pathExists(packageJsonPath)) {
    return {};
  }
  let { dependencies, devDependencies, peerDependencies } = JSON.parse(await fs__default$1.readFile(packageJsonPath, "utf8"));
  const allDependencies = {
    ...dependencies,
    ...devDependencies,
    ...peerDependencies
  };
  return Object.fromEntries(Object.entries(allDependencies).map(([name, version]) => {
    let majorVersion;
    if (typeof version !== "string") {
      majorVersion = 0;
    } else if (version.startsWith("^") || version.startsWith("~")) {
      majorVersion = parseInt(version.slice(1));
    } else {
      majorVersion = parseInt(version);
    }
    const readInstalledVersion = async () => {
      try {
        const moduleEntryPath = require.resolve(name, {
          paths: [rootDirPath]
        });
        let packagePath = moduleEntryPath;
        let packageJsonPath2 = null;
        while (packagePath !== path__default.dirname(packagePath)) {
          const candidatePackageJsonPath = path__default.join(packagePath, "package.json");
          if (fs__default$1.existsSync(candidatePackageJsonPath)) {
            packageJsonPath2 = candidatePackageJsonPath;
            break;
          }
          packagePath = path__default.dirname(packagePath);
        }
        if (!packageJsonPath2) {
          throw new Error(`No package.json path found from: ${moduleEntryPath}`);
        }
        const packageInfo = JSON.parse(await fs__default$1.readFile(packageJsonPath2, "utf8"));
        const version2 = packageInfo["version"];
        if (!version2 || typeof version2 !== "string") {
          throw new Error(`Invalid version found for package: ${packageJsonPath2}`);
        }
        return version2;
      } catch (e) {
        console.error(`Unable to read installed version of package: ${name}`, e);
        return null;
      }
    };
    return [name, { majorVersion, readInstalledVersion }];
  }));
}

async function createWorkspace({
  versionCode,
  rootDirPath,
  reader,
  frameworkPlugin,
  logLevel,
  middlewares,
  onReady,
  persistedStateManager = new LocalFilePersistedStateManager()
}) {
  const expectedPluginApiVersion = 3;
  if (!frameworkPlugin.pluginApiVersion || frameworkPlugin.pluginApiVersion < expectedPluginApiVersion) {
    throw new Error(`Preview.js framework plugin ${frameworkPlugin.name} is incompatible with this version of Preview.js. Please upgrade it.`);
  } else if (frameworkPlugin.pluginApiVersion > expectedPluginApiVersion) {
    throw new Error(`Preview.js framework plugin ${frameworkPlugin.name} is too recent. Please upgrade Preview.js or use an older version of ${frameworkPlugin.name}.`);
  }
  if (frameworkPlugin.transformReader) {
    reader = frameworkPlugin.transformReader(reader, rootDirPath);
  }
  const collected = {};
  const typeAnalyzer$1 = typeAnalyzer.createTypeAnalyzer({
    reader,
    rootDirPath,
    collected,
    specialTypes: frameworkPlugin.specialTypes,
    tsCompilerOptions: frameworkPlugin.tsCompilerOptions
  });
  const router = new ApiRouter();
  router.registerEndpoint(api.localEndpoints.GetInfo, async () => {
    const separatorPosition = versionCode.indexOf("-");
    if (separatorPosition === -1) {
      throw new Error(`Unsupported version code format: ${versionCode}`);
    }
    const platform = versionCode.substr(0, separatorPosition);
    const version = versionCode.substr(separatorPosition + 1);
    return {
      appInfo: {
        platform,
        version
      }
    };
  });
  router.registerEndpoint(api.localEndpoints.GetState, persistedStateManager.get);
  router.registerEndpoint(api.localEndpoints.UpdateState, persistedStateManager.update);
  router.registerEndpoint(api.localEndpoints.ComputeProps, async ({ filePath, componentName }) => {
    const component = (await frameworkPlugin.detectComponents(typeAnalyzer$1, [
      path__default.join(rootDirPath, filePath)
    ])).find((c) => c.name === componentName);
    if (!component) {
      return {
        types: {
          props: typeAnalyzer.UNKNOWN_TYPE,
          all: {}
        },
        args: []
      };
    }
    return computeProps({
      component
    });
  });
  const previewer = new Previewer({
    reader,
    rootDirPath,
    previewDirPath: path__default.join(path__default.dirname(path__default.dirname(require.resolve("@previewjs/iframe"))), "preview"),
    frameworkPlugin,
    logLevel,
    middlewares: [
      express__default.json(),
      cookieParser__default(),
      express__default.Router().use("/monaco-editor", express__default.static(path__default.join(__dirname, "monaco-editor"))),
      async (req, res, next) => {
        if (req.path.startsWith("/api/")) {
          res.json(await router.handle(req.path.substr(5), req.body, req, res));
        } else {
          next();
        }
      },
      ...middlewares
    ],
    onFileChanged: (absoluteFilePath) => {
      const filePath = path__default.relative(rootDirPath, absoluteFilePath);
      for (const name of Object.keys(collected)) {
        if (name.startsWith(`${filePath}:`)) {
          delete collected[name];
        }
      }
    }
  });
  const workspace = {
    rootDirPath,
    reader,
    typeAnalyzer: typeAnalyzer$1,
    frameworkPlugin,
    preview: {
      start: async (allocatePort) => {
        const port = await previewer.start(async () => {
          const port2 = allocatePort ? await allocatePort() : 0;
          return port2 || await getPort__default({
            port: getPort__default.makeRange(3140, 4e3)
          });
        });
        return {
          url: () => `http://localhost:${port}`,
          stop: async (options) => {
            await previewer.stop(options);
          }
        };
      }
    },
    components: {
      list: (options) => {
        return analyzeProject(workspace, options);
      }
    },
    dispose: async () => {
      typeAnalyzer$1.dispose();
    }
  };
  if (onReady) {
    await onReady({
      registerEndpoint: (endpoint, handler) => router.registerEndpoint(endpoint, handler),
      workspace
    });
  }
  return workspace;
}
function findWorkspaceRoot(absoluteFilePath) {
  let dirPath = path__default.resolve(absoluteFilePath);
  while (dirPath !== path__default.dirname(dirPath)) {
    if (fs__default$1.existsSync(path__default.join(dirPath, "package.json"))) {
      return dirPath;
    }
    dirPath = path__default.dirname(dirPath);
  }
  return null;
}

exports.createWorkspace = createWorkspace;
exports.findWorkspaceRoot = findWorkspaceRoot;
exports.generateComponentId = generateComponentId;
exports.loadPreviewEnv = loadPreviewEnv;
